#include "Panel_dynamics.h"

#include "Program_title.h"
#include "MD_simulator.h"
#include "Simulator_factory.h"
#include "Project.h"
#include "Initialization_stage.h"
#include "Timer.h"
#include "Thermostat.h"
#include "Model_kit.h"
#include "Lock.h"
#include "Mach_eps.h"
#include "Profiler.h"

#include "Log.h"
#include "User.h"
#include "Main_window.h"

#include <QCloseEvent>

namespace MM
{   

Panel_dynamics::
Panel_dynamics ()
:
    Panel ("Dynamics", "left"),
    player_(Main_window::singleton())
{
    setupUi (this);

#ifndef ASCALAPH
    Tool->removeItem (Tool->findText("MDynaMix"));
#endif

    pushButton_Options->hide();
    stackedWidget_Thermostat->hide();
    comboBox_Thermostat->hide();

    connect(Tool,           SIGNAL(currentIndexChanged (const QString &)),
            this,           SLOT(set_tool (const QString &)));


    connect(checkBox_Barostat,          SIGNAL(stateChanged (int)),
            this,                       SLOT(barostat     (int)));



    connect(checkBox_Thermostat,        SIGNAL(stateChanged (int)),
            this,                       SLOT(thermostat     (int)));

    connect (comboBox_Thermostat, SIGNAL(currentIndexChanged(const QString&)),
             this,                SLOT  (show_thermostat    (const QString&)));


    
    connect(doubleSpinBox_Temperature,  SIGNAL(valueChanged (double)),
            this,                       SLOT(set_temperature(double)));
       

    connect(doubleSpinBox_Time_per_long_step,  SIGNAL(valueChanged (double)),
            this,                       SLOT(set_steps(double)));
       
    connect(doubleSpinBox_Duration,     SIGNAL(valueChanged (double)),
            this,                       SLOT(set_steps(double)));
       
    renew ();
}

void Panel_dynamics::
set_stop_state ()
{
    player_->player_on_ = false;

    player_->toolButton_Stop   -> setEnabled (false);
    player_->toolButton_Tune   -> setEnabled (true);
    player_->toolButton_Run    -> setEnabled (true);
    player_->toolButton_Start  -> setEnabled (true);

    if (this->isHidden ())
        turn (false);
}

void Panel_dynamics::
set_run_state ()
{
    player_->player_on_ = true;

    player_->toolButton_Stop   -> setEnabled (true);
    player_->toolButton_Tune   -> setEnabled (false);
    player_->toolButton_Run    -> setEnabled (false);
    player_->toolButton_Start  -> setEnabled (false);
}

void Panel_dynamics::
turn (bool on)
{
    //MD_simulator & simulator = model().kit().current_MD();

    //if (!simulator.is_running())
    if (!player_->player_on_)
    {
        player_->toolButton_Stop   -> disconnect ();
        player_->toolButton_Back   -> disconnect ();
        player_->toolButton_Tune   -> disconnect ();
        player_->toolButton_Run    -> disconnect ();
        player_->toolButton_Start  -> disconnect ();
        player_->toolButton_Forward-> disconnect ();
        player_->toolButton_Pause  -> disconnect ();

        if (on)
        {
            player_->tabWidget_bottom->setCurrentWidget (player_->tab_Run);

            connect (player_->toolButton_Stop,   SIGNAL(clicked()), this,
                                 SLOT  (stop   ()));
            connect (player_->toolButton_Tune,   SIGNAL(clicked()), this,
                                 SLOT  (tune   ()));
            connect (player_->toolButton_Run,    SIGNAL(clicked()), this,
                                 SLOT  (run    ()));
            connect (player_->toolButton_Start,  SIGNAL(clicked()), this,
                                 SLOT  (start  ()));
            set_stop_state ();
        }
        else 
        {
            player_->toolButton_Stop   -> setEnabled (false);
            player_->toolButton_Back   -> setEnabled (false);
            player_->toolButton_Tune   -> setEnabled (false);
            player_->toolButton_Run    -> setEnabled (false);
            player_->toolButton_Start  -> setEnabled (false);
            player_->toolButton_Forward-> setEnabled (false);
            player_->toolButton_Pause  -> setEnabled (false);
        }
    }

    if (on)
        renew ();
}

void Panel_dynamics::
closeEvent (QCloseEvent * event)
{
    MD_simulator & simulator = model().kit().current_MD();

    if (simulator.is_running())
        event->ignore ();
}

void Panel_dynamics::
set_tool (const QString & text)
{
    //Text tool(text);

    //if (text == "Ascalaph")
    //    tool = "MD";

    //MD_simulator::adopt (dynamic_cast <MD_simulator*> 
    //    (Simulator_factory::create (tool)));

    if (text == "Ascalaph MD")
        model().kit().set_MD (model().kit().MD ());

    #ifdef ASCALAPH
    else if (text == "MDynaMix")
        model().kit().set_MD (model().kit().MDynaMix ());
    #endif

    renew ();
}

void Panel_dynamics::
barostat (int on)
{
    doubleSpinBox_Pressure->setEnabled   (on);
    model().kit().current_MD().set_barostat (on);
}

void Panel_dynamics::
thermostat (int on)
{
    doubleSpinBox_Temperature->setEnabled    (on);
    model().kit().current_MD().set_thermostat (on);
}

void Panel_dynamics::
set_temperature (double value)
{
    Temperature_unit t;
    t.set_K (value);
    model().kit().current_MD().set_desired_temperature (t);
}

void Panel_dynamics::
set_steps (double)
{
    Duration_unit duration;
    duration.set_step_length_ps (doubleSpinBox_Time_per_long_step->value()/1000.);
    duration.set_ps             (doubleSpinBox_Duration->value());
    model().kit().current_MD().set_duration (duration);
}

void Panel_dynamics::
renew_current_model (Model &)
{
    renew ();
}

void Panel_dynamics::
renew (Model_event::Hint /*hint*/)
{
    MD_simulator & simulator = model().kit().current_MD();


    Text const & tool_name = model().kit().current_MD().title();

    if (tool_name == "MD")
        Tool->setCurrentIndex (Tool->findText("Ascalaph MD"));

    else if (tool_name == "MDynaMix")
        Tool->setCurrentIndex (Tool->findText("MDynaMix"));
    else
        FLAW (tool_name + " - no such tool.");


    bool support_barostat = simulator.support_barostat();

    if (support_barostat)
    {
        groupBox_Barostat->setEnabled (true);
        doubleSpinBox_Pressure->setValue (simulator.desired_pressure().atm());

        bool has_barostat = simulator.has_barostat();

        checkBox_Barostat->setChecked (has_barostat);

        if (has_barostat)
        {
            doubleSpinBox_Pressure->setEnabled (true);
            checkBox_Barostat->setChecked (true);
        }
        else
        {
            doubleSpinBox_Pressure->setEnabled (false);
            checkBox_Barostat->setChecked (false);
        }

        doubleSpinBox_Pressure->setValue (simulator.desired_pressure().atm());
   }
    else
        groupBox_Barostat->setEnabled (false);


    bool has_thermostat = simulator.has_thermostat();

    checkBox_Thermostat->setCheckState 
        (has_thermostat ? Qt::Checked : Qt::Unchecked);

    doubleSpinBox_Temperature->setEnabled    (has_thermostat);

    if (has_thermostat)
    {
///        Text type = MD_simulator::current().thermostat().title ();
        //to_user().status (type);
        doubleSpinBox_Temperature->setValue (simulator.desired_temperature ().K());
    }

    //fix to unit viewer
    doubleSpinBox_Temperature->setValue 
        (simulator.desired_temperature().K());

    doubleSpinBox_Time_per_long_step->blockSignals (true);
    doubleSpinBox_Duration->blockSignals (true);

    Duration_unit duration = simulator.duration();
    doubleSpinBox_Time_per_long_step->setValue (duration.step_length_ps() * 1000.);
    doubleSpinBox_Duration->          setValue (duration.ps() + d_mach_eps_2);

    doubleSpinBox_Duration->blockSignals (false);
    doubleSpinBox_Time_per_long_step->blockSignals (false);
}

void Panel_dynamics::
show_thermostat (const QString & type)
{
    for (int i=0;  i<stackedWidget_Thermostat->count();  ++i)
    {
        if (stackedWidget_Thermostat->widget(i)->objectName() == type)
            stackedWidget_Thermostat->setCurrentIndex (i);
    }
}


void Panel_dynamics::
run ()
{
    tune  ();
    start ();
}

void Panel_dynamics::
tune ()
{
    MD_simulator & simulator = model().kit().current_MD();
    //Model        & model     = Project::singleton().current_model();//fix
    
    //simulator.set_model (model());//fix
    
    //simulator.set_steps (doubleSpinBox_Duration->value());  //fix to connect

    struct Process_events : public Command
    {
        int i_;

        Process_events () : i_(0) {}

        void execute () 
        {
            if (++i_ % 1 == 0)
                //qApp->processEvents (QEventLoop::AllEvents, 100);
                qApp->processEvents(QEventLoop::ExcludeSocketNotifiers);
        }
    };
    simulator.after_each_step (new Process_events);

    if (checkBox_Renew_velocities->checkState() == Qt::Checked)
    {
        Temperature_unit t;
        t.set_K (doubleSpinBox_Temperature->value());
        Initialization_stage::setup_velocities (model(), t);
    }

    model().kit().strip (true);
    simulator.tune ();

    to_user().status ("Ready to run");
}

MD_simulator * simulator = 0;

void Panel_dynamics::
start ()
{
    MD_simulator & simulator = model().kit().current_MD();
    MM::simulator = & simulator;
    Timer timer;
    Text  time;

    //fix 
    //model().kit().interaction ().set_lock (true);
    Lock <Interaction> lock (model().kit().interaction ());

    to_user().status ("Started");

    profile().restart();

    set_run_state ();
    
    try
    {
        simulator.start ();
    }
    catch(Text & message)   {stop (); to_user().error (message);}
    catch(Flaw & exception) {stop (); to_user().error (exception.message());}
    catch(Contract_violation & exc) {stop (); to_user().error (exc.message());}
    catch(...)              {stop (); to_user().fatal_error ("Unknown error.");}

    set_stop_state ();

    //model().kit().interaction ().report (&time, "F speed");
    profile().report();

    time += "run ";
    time += timer.elapsed();    timer.restart();
    time += " sec\n";
    log () << time;
}

void Panel_dynamics::
stop ()
{
    if (MM::simulator != 0)
        MM::simulator->stop ();
    //MM::simulator = 0;

    to_user ().info ("Simulation was stopped by the user. "
        "Please wait for the end of trajectory reading.");
}

}//MM
