#include "MD.h"

#include "Defs.h"
#include "Coordinates.h"
#include "Velocities.h"
#include "Force.h"
#include "Mass.h"
#include "MD_method.h"
#include "Thermostat.h"
//#include "MD_subject.h"
#include "Timer.h"

#include "Log.h"

#include <math.h>

namespace MM
{

const int max_descriptions = 10;//fix

MD::
MD ()
:   
    step_length_(1e-015 * Constant::sec_to_time_unit),
    drift_length_(0.),
    kick_length_(0.),
    subject_ (0),
    method_(0),
    thermostat_(&Thermostat::none()),
    stop_flag_(false)
{
}

MD::
MD (MD_subject & subject, MD_method & method)
:   
    step_length_(1e-015 * Constant::sec_to_time_unit),
    drift_length_(0.),
    kick_length_(0.),
    subject_ (&subject),
    method_(&method),
    thermostat_(&Thermostat::none()),
    stop_flag_(false)
{
    method_->set_simulation (*this);
}


Text MD::
title () const
{
    return Text() + "MD: " + method_->title();
}

double MD::
step_length () const
{
    return step_length_;
}

void MD::
set_step_length (double new_length)
{
    step_length_ = new_length;
}

void MD::
set_method (MD_method & new_method)
{
    method_ = &new_method;
    method_->set_simulation (*this);
}

void MD::
set_subject (MD_subject & new_subject)
{
    subject_ = &new_subject;
}

void MD::
start ()
{
    algorithm().description (Text("start  ") + title());
    stop_flag_ = false;
    method_->start();
}

void MD::
cycle ()
{
    algorithm().description (Text("begin ") + iterations() + " step cycle");

    Timer timer;    Text  time;
    int steps = iterations();

    for (int i=0;  i<steps && !stop_flag_;  ++i)
    {
        algorithm().description (Text(i));

        method_->step ();

        thermostat_->execute(); 
        step_finished_.publish ();

        //if (timer.elapsed() > .05)
        {
            update ();                      //fix
            timer.restart();
        }

        if (i >= max_descriptions)
            algorithm().off();

        //time += "iteration ";
        //time += timer.elapsed();    
        //time += " sec\n";
        //log () << time;
    }
    algorithm().description ("end cycle");
}

int MD::
iterations () const
{
    return method_->iterations ();
}

void MD::
finish ()
{
    method_->finish ();
    stop_.publish ();
    algorithm().description (Text("finish ") + title());
    ENSURE ("", fabs(drift_length_ - kick_length_) < 0.000001 * drift_length_);
}

void MD::
F (Hint hint)
{
    //Timer timer;    Text  time;

    subject_->F (hint);
    algorithm().description (Text("F       ") + hint);

    //time += "F         ";
    //time += timer.elapsed();    timer.restart();
    //time += " sec\n";
    //log () << time;
}

void MD::
kick (double h)
{
    //Timer timer;    Text  time;

    kick_length_ += h * step_length_;
    subject_->kick (h*step_length_);
    algorithm().description
        (Text("kick    ") + h*step_length_ + "  \tsum = " + kick_length_);

    //time += "kick      ";
    //time += timer.elapsed();    timer.restart();
    //time += " sec\n";
    //log () << time;
}

void MD::
drift (double h)
{
    //Timer timer;    Text  time;

    drift_length_ += h * step_length_;
    subject_->drift (h*step_length_);
    algorithm().description
        (Text("drift   ") + h*step_length_ + "  \tsum = " +drift_length_);

    //time += "drift     ";
    //time += timer.elapsed();    timer.restart();
    //time += " sec\n";
    //log () << time;
}

void MD::
update ()
{
    subject_->update ();
}

/*
bool MD::
invariant () const
{
    bool result = coordinats_ != 0 &&
                  velocities_ != 0 &&
                  forces_     != 0 &&
                  masses_     != 0;
    if (!result)    
        return false;

    int coordinats = coordinats_-> size ();
    int velocities = velocities_-> size ();
    int forces     = forces_->     size ();
    int masses     = masses_->     size ();

    result = coordinats == velocities &&
             coordinats == forces &&
             coordinats == masses;

    return result;
}//*/

}//MM









